
// pubcrawl version 2022.08.05
// https://yx7.cc/code/

#include <iostream>
#include <iomanip>
#include <sstream>

#include <numeric>

#include <optional>
#include <variant>
#include <array>
#include <vector>

#include <queue>

#include <unordered_set>
#include <map>
#include <unordered_map>

#include <utility>
#include <algorithm>

#include <thread>
#include <future>
#include <mutex>

#include <cassert>
#include <csignal>

#include <sys/resource.h>

const unsigned num_threads = 1 + std::thread::hardware_concurrency();

////////////////////////////////////////////////////////////////

#define NUM_PRIMES 74
#define PUB_BYTES 64

extern "C" {
    bool csidh(unsigned char *out, unsigned char const *in, int8_t const *priv);
}

////////////////////////////////////////////////////////////////

class vec
{
public:
    vec() : e{0} { }

    int8_t const *data() const { return this->e.data(); }
    int8_t& operator[](size_t i) { return std::ref(this->e[i]); }
    int8_t const& operator[](size_t i) const { return this->e[i]; }

    vec operator-() const
    {
        vec u;
        for (size_t i = 0; i < NUM_PRIMES; ++i)
            u[i] = -(*this)[i];
        return u;
    }

    vec operator+(vec const& v) const
    {
        vec u;
        for (size_t i = 0; i < NUM_PRIMES; ++i)
            u[i] = (*this)[i] + v[i];
        return u;
    }

    vec operator-(vec const& v) const
    {
        vec u;
        for (size_t i = 0; i < NUM_PRIMES; ++i)
            u[i] = (*this)[i] - v[i];
        return u;
    }

    bool operator==(vec const& v) const { return this->e == v.e; }
    bool operator!=(vec const& v) const { return this->e != v.e; }

    uint64_t norm1() const
    {
        return std::accumulate(this->e.begin(), this->e.end(), 0,
                               [](uint64_t a, int64_t e) { return a + std::abs(e); });
    }

private:
    std::array<int8_t,NUM_PRIMES> e;
};

class pub
{
public:
    uint8_t *data() { return this->A.data(); }
    uint8_t const *data() const { return this->A.data(); }
    pub() : A{0} { }

    bool operator==(pub const& E) const { return this->A == E.A; }
    bool operator!=(pub const& E) const { return this->A != E.A; }

private:
    std::array<uint8_t,PUB_BYTES> A;
};

pub operator*(vec const& v, pub const& E)
{
    pub EE;
    csidh(EE.data(), E.data(), v.data());
    return EE;
}

std::ostream& operator<<(std::ostream& o, vec const& v)
{
    std::ostringstream s;
#if 0
    for (size_t i = 0; i < NUM_PRIMES; ++i)
        s << (v[i] > 0 ? '+' : v[i] < 0 ? '-' : ' ') << std::abs(v[i]);
#else
    s << "{";
    bool first = true;
    for (size_t i = 0; i < NUM_PRIMES; ++i) {
        if (v[i]) {
            s << (first ? "" : ",") << i << ":" << (int) v[i];
            first = false;
        }
    }
    s << "}";
#endif
    return o << s.str();
}

std::ostream& operator<<(std::ostream& os, pub const& E)
{
    uint8_t const *A = E.data();
    std::ostringstream s;
    s << std::hex << std::setfill('0');
    for (size_t i = PUB_BYTES-1; i < PUB_BYTES; --i)
        s << std::setw(2) << (unsigned) A[i];
    return os << s.str();
}

std::istream& operator>>(std::istream& is, pub& E)
{
    uint8_t *A = E.data();
    std::string s(3, '\0');
    for (size_t i = PUB_BYTES-1; i < PUB_BYTES; --i) {
        if (!is.get(s.data(), s.size()))
            break;
        A[i] = std::stoi(s, nullptr, 16);
    }
    return is;
}

using path2 = std::pair<pub,vec>;
using path3 = std::tuple<pub,vec,pub>;
using path4 = std::tuple<pub,vec,pub,vec>;

namespace std {
    template<>
    struct hash<pub>
    {
        size_t operator()(pub const& E) const {
            std::string_view s {reinterpret_cast<char const *>(E.data()), PUB_BYTES};
            return std::hash<std::string_view>{}(s);
        }
    };

    template<>
    struct hash<path2>
    {
        size_t operator()(path2 const& tup) const {
            auto [E,v] = tup;
            std::string_view s {reinterpret_cast<char const *>(v.data()), NUM_PRIMES};
            return 1111111*std::hash<pub>()(E) ^ std::hash<std::string_view>{}(s);
        }
    };
}

////////////////////////////////////////////////////////////////

std::atomic<bool> done = false;
std::atomic<unsigned> waiting = 0;

std::mutex mtx_io;

/* configuration: which isogeny steps to use */
std::vector<vec> step(NUM_PRIMES);      // i -> "secret" vector to compute i'th step
std::array<bool,NUM_PRIMES> support;    // i -> whether to use the i'th step at all
std::unordered_map<pub,int> signs;      // E -> walk only + or -; omitted means both
unsigned max_dist = -1;                 // limit # of isogeny steps from each node

/* work queue */
std::mutex mtx_sync;
std::deque<std::variant<path3,path4>> todo;
std::condition_variable cv;

/* lookup table: paths to all nodes ever seen */
std::mutex mtx_tab;
std::unordered_map<pub, path2> tab;
std::unordered_map<uint64_t,size_t> norm_counts;

/* connected components */
std::mutex mtx_comps;
std::vector<std::unordered_set<pub>> comps;

////////////////////////////////////////////////////////////////

void set_done()
{
    {
        std::lock_guard lock(mtx_sync);
        done = true;
    }
    cv.notify_all();
}

std::atomic<uint64_t> csidhs_done = 0;

void status_thread()
{
    constexpr std::chrono::seconds status_interval(1);

    uint64_t last_csidhs_done = 0;

    auto const start_time = std::chrono::steady_clock::now();

    do {

        auto const last_time = std::chrono::steady_clock::now();

        do {
            auto elapsed = std::chrono::steady_clock::now() - last_time;
            if (elapsed >= status_interval)
                break;
            std::unique_lock lock(mtx_sync);
            cv.wait_for(lock, status_interval - elapsed);
        } while (!done);

        auto const cur_time = std::chrono::steady_clock::now();

        size_t csz, tsz, qsz;
        std::map<uint64_t,size_t> cnts;
        { std::lock_guard lock(mtx_comps);  csz = comps.size();                                     }
        { std::lock_guard lock(mtx_tab);    tsz = tab.size();
                                            cnts.insert(norm_counts.begin(), norm_counts.end());    }
        { std::lock_guard lock(mtx_sync);   qsz = todo.size();                                      }

        std::ostringstream plens;
        if (!cnts.empty()) {
            uint64_t norm = 0;
            for (auto const& cnt: cnts) {
                for ( ; norm < cnt.first; ++norm)
                    plens << ",";
                plens << "\x1b[34m" << cnt.second << "\x1b[0m";
            }
        }

        {
            auto const tot_seconds = 1e-9 + std::chrono::duration_cast<std::chrono::duration<double>>(cur_time - start_time).count();
            auto const cur_seconds = 1e-9 + std::chrono::duration_cast<std::chrono::duration<double>>(cur_time - last_time).count();

            std::ostringstream ss;
            ss << "\x1b[K"
                      << "components left: \x1b[36m" << csz << "\x1b[0m"
                      << " | "
                      << "table size: \x1b[35m" << tsz << "\x1b[0m"
                      << "(" << plens.str() << ")"
                      << " | "
                      << "queue length:"
                      << " \x1b[35m" << qsz << "\x1b[0m"
                      << " | "
                      << "steps/sec:"
                      << " cur:\x1b[35m" << (uint64_t) ((csidhs_done - last_csidhs_done) / cur_seconds) << "\x1b[0m"
                      << " avg:\x1b[35m" << (uint64_t) (csidhs_done / tot_seconds) << "\x1b[0m"
                      << "\r";

            std::lock_guard lock(mtx_io);
            std::cerr << ss.str() << std::flush;
        }

        last_csidhs_done = csidhs_done;
    } while (!done);

    std::cerr << "\n";
}

////////////////////////////////////////////////////////////////

std::vector<std::unordered_set<pub>>::iterator component_of(pub const& E)
{
    for (auto it = comps.begin(); it != comps.end(); ++it)
        if (it->find(E) != it->end())
            return it;
    return comps.end();
}

std::vector<path4> neighbors(path3 const& item)
{
    std::vector<path4> res;

    auto const &[E0,d,E] = item;
    assert(d.norm1() <= max_dist);

    int sign = 0;
    if (auto it = signs.find(E0); it != signs.end())
        sign = it->second;

    for (size_t i = 0; i < NUM_PRIMES; ++i) {
        if (!support[i])
            continue;

        if (d[i]) {  // avoid backtracking
                res.emplace_back(E0, d, E, (d[i] > 0 ? step[i] : -step[i]));
                break;  // stop at the first non-zero entry to make sure each node is only visited once
        }

        if (sign >= 0)
            res.emplace_back(E0, d, E,  step[i]);
        if (sign <= 0)
            res.emplace_back(E0, d, E, -step[i]);
    }

    return res;
}

#define move_it std::make_move_iterator

void expand_front()     // mtx_sync
{
    while (!todo.empty()) {

        auto const& item = todo.front();
        if (std::holds_alternative<path4>(item))
            return;
        assert(std::holds_alternative<path3>(item));

        std::vector<path4> new_todo;

        auto neighs = neighbors(std::move(std::get<path3>(item)));
        todo.pop_front();
        for (auto const& neigh: neighs) {
            auto const& [E0,v,_,w] = neigh;
            if ((v+w).norm1() > max_dist)
                continue;
            new_todo.push_back(std::move(neigh));
        }

        todo.insert(todo.begin(), move_it(new_todo.begin()), move_it(new_todo.end()));
    }
}

#define BATCH 100

void grow()
{
    while (!done) {
        std::vector<path4> work;
        {
            std::unique_lock lock(mtx_sync);

            if (todo.empty()) {
                if (++waiting >= num_threads) {
                    lock.unlock();
                    set_done();
                    return;
                }
                if (done)
                    return;
                cv.wait(lock);
                --waiting;
            }

            while (work.size() < BATCH) {

                expand_front();
                if (todo.empty())
                    break;

                auto item = std::move(todo.front());
                todo.pop_front();

                assert(std::holds_alternative<path4>(item));
                work.push_back(std::move(std::get<path4>(item)));
            }
        }

//{
//std::lock_guard lock(mtx_io);
//std::cerr << "thread " << std::this_thread::get_id() << " work units: " << work.size() << std::endl;
//}

        if (work.empty())
            continue;

        std::vector<pub> results;
        results.reserve(work.size());
        for (auto const& [E0,v,E,w]: work) {
//{
//std::lock_guard lock(mtx_io);
//std::cerr << "! " << E0 << " " << " " << (v+w).norm1() << " " << v+w << std::endl;
//}
            if (done)
                return;
            results.push_back(w * E);   // slow
        }
        csidhs_done += results.size();
        assert(results.size() == work.size());

        std::vector<path3> paths, new_todo;
        paths.reserve(work.size());
        new_todo.reserve(work.size());
        {
            std::lock_guard lock(mtx_tab);

            for (size_t i = 0; i < work.size(); ++i) {
                if (done)
                    return;

                auto const &[E0,v,E,w] = work[i];
                assert(v*E0 == E);

                pub EE = results[i];
                assert(w*E == EE);

//{
//std::lock_guard io(mtx_io);
//std::cerr << E0 << " " << EE << " " << v+w << std::endl;
//}

                auto [it,ins] = tab.emplace(EE, path2(E0, v+w));
                if (ins) {  // previously unseen
                    ++norm_counts[(v+w).norm1()];
                    if ((v+w).norm1() >= max_dist)
                        continue;
                    new_todo.emplace(new_todo.end(), E0, v+w, EE);
                }
                else {
                    auto const &[E1,x] = it->second;
                    paths.emplace_back(E0, v+w - x, E1);
                }
            }
        }

        {
            std::lock_guard lock(mtx_comps);

            for (auto const& [E0,d,E1]: paths) {
                if (auto c0 = component_of(E0), c1 = component_of(E1); c0 != c1) {
                    c0->merge(*c1);
                    comps.erase(c1);
                    {
                        std::lock_guard lock(mtx_io);
                        std::cerr << "\x1b[32m" << E0 << " " << E1 << " " << d << "\x1b[0m" << std::endl;
                        std::cout << E0 << " " << E1 << " " << d << std::endl;
                    }
                }
            }

            if (comps.size() <= 1)
                set_done();
        }

        {
            std::lock_guard lock(mtx_sync);
            todo.insert(todo.end(), move_it(new_todo.begin()), move_it(new_todo.end()));
        }
        cv.notify_all();
    }
}

////////////////////////////////////////////////////////////////

void handler(int num)
{
    (void) num;
    set_done();
}

__attribute__((noreturn))
void bad_args()
{
    std::cerr << "\x1b[31mbad arguments\x1b[0m\n";
    exit(-1);
}

template <class It>
void explode(std::string const& str, char sep, It out)
{
    std::istringstream buf(str);
    std::string tmp;
    while (std::getline(buf, tmp, sep))
        *out++ = tmp;
}

int main(int argc, char **argv)
{
    std::vector<pub> pubs;
    while (std::cin) {
        std::string l;
        if (!std::getline(std::cin, l))
            break;
        pub E;
        if (!(std::istringstream(l) >> E))
            break;

        if (std::find(pubs.begin(), pubs.end(), E) == pubs.end())   // deduplicate
            pubs.push_back(E);
    }
std::cerr << "loaded " << pubs.size() << " distinct public keys" << std::endl;

    if (pubs.size() < 2) {
        std::cerr << "\x1b[31mnot enough target curves\x1b[0m\n";
        exit(1);
    }

////////////////////////////////////////////////////////////////

    std::unordered_map<std::string, std::string> args;
    for (ssize_t i = 1; i < argc; ) {
        std::string a(argv[i++]);
        if (a.empty() || a[0] != '-' || i >= argc || args.find(a) != args.end())
            bad_args();
        args[a] = argv[i++];
    }

    if (auto it = args.find("--sign"); it != args.end()) {
        int sign;
        if (it->second == "+")      sign = +1;
        else if (it->second == "-") sign = -1;
        else bad_args();
        signs[pubs.front()] = sign;
        signs[pubs.back()] = -sign;
    }

    if (auto it = args.find("--support"); it != args.end()) {
        std::vector<std::string> ranges;
        explode(it->second, ',', std::back_inserter(ranges));
        support.fill(false);
        for (auto const& range: ranges) {
            std::vector<std::string> bounds;
            explode(range, '-', std::back_inserter(bounds));
            unsigned long lo = NUM_PRIMES, hi = NUM_PRIMES;
            if (bounds.size() == 1) {
                hi = lo = std::stoul(bounds[0]);
            }
            else if (bounds.size() == 2) {
                lo = std::stoul(bounds[0]);
                hi = std::stoul(bounds[1]);
            }
            if (lo > hi || hi >= NUM_PRIMES)
                bad_args();
            std::fill_n(support.begin() + lo, hi-lo+1, true);
        }
    }
    else {
        support.fill(true);
    }

    unsigned step_mult = 2;
    if (auto it = args.find("--mult"); it != args.end())
        step_mult = std::stoull(it->second);
    for (size_t i = 0; i < NUM_PRIMES; ++i)
        step[i][i] = step_mult;

    if (auto it = args.find("--dist"); it != args.end())
        max_dist = std::stoull(it->second);

////////////////////////////////////////////////////////////////

    signal(SIGINT, &handler);
//    setpriority(PRIO_PGRP, 0, NZERO-1);     // reduce priority to avoid hanging

////////////////////////////////////////////////////////////////

    {
        comps.reserve(pubs.size());
        tab.reserve((size_t) 1 << 22);      // or something
        for (auto const& E: pubs) {
            comps.push_back({E});
            todo.push_back(path3(E, {}, E));
            tab.emplace(E, path2(E, {}));
            ++norm_counts[0];
std::cout << E << "\n";
        }
    }

    std::vector<std::thread> threads;
    while (threads.size() < num_threads)
        threads.emplace(threads.end(), grow);

    threads.emplace(threads.end(), status_thread);

    for (auto& t: threads)
        t.join();
}

